datablock ItemData(GravityGunItem) {
	shapeFile = "./shapes/gravity_gun.dts";
	emap = true;

	mass = 1;
	density = 0.2;
	friction = 0.6;
	elasticity = 0.2;

	uiName = "Gravity Gun";
	iconName = "";

	doColorShift = 1;
	colorShiftColor = "1 0 0 1";

	image = GravityGunImage;
	canDrop = true;
};

datablock ShapeBaseImageData(GravityGunImage) {
	shapeFile = GravityGunItem.shapeFile;
	emap = GravityGunItem.emap;

	mountPoint = 0;
	offset = "0 0 0";

	correctMuzzleVector = 0;
	armReady = 1;

	item = GravityGunItem;
	ammo = " ";

	doColorShift = GravityGunItem.doColorShift;
	colorShiftColor = GravityGunItem.colorShiftColor;

	stateName[0] = "Activate";
	stateTimeoutValue[0] = 0.015;
	stateTransitionOnTimeout[0] = "Ready";

	stateName[1] = "Ready";
	stateAllowImageChange[1] = 1;
	stateTransitionOnTriggerDown[1] = "Fire";

	stateName[2] = "Fire";
	stateScript[2] = "onPrimaryFire";
	stateTransitionOnTriggerUp[2] = "Ready";
};

function GravityGunImage::onUnMount(%this, %obj, %slot) {
	if (isObject(%obj.gravityGunTarget)) {
		%obj.gravityGunTarget = "";
		cancel(%obj.gravityGunTargetUpdate);
	}
}

// because port is old and can't remember:
// this flings aimed at object or held object

function GravityGunImage::onPrimaryFire(%this, %obj, %slot) {
	if (isObject(%obj.gravityGunTarget)) {
		%col = %obj.gravityGunTarget;
		%obj.gravityGunTarget = "";

		cancel(%obj.gravityGunTargetUpdate);
	}
	else {
		%col = %this.getTargetObject(%obj, %slot, 25);

		if (isObject(%col) && (%col.getType() & $TypeMasks::PlayerObjectType)) {
			%mount = %col.getObjectMount();

			if (isObject(%mount)) {
				%col = %mount;
			}
		}
	}

	if (!isObject(%col)) {
		return;
	}

	if (%col.getType() & $TypeMasks::StaticShapeObjectType) {
		return;
	}

	if (%col.getType() & $TypeMasks::FxBrickObjectType) {
		return;
	}

	if (%col.getType() & $TypeMasks::TerrainObjectType) {
		return;
	}

	%point = %obj.getForwardVector();
	%vector = %obj.getMuzzleVector(%slot);

	%directionModifier = 0.75;

	%minForce = 24;
	%maxForce = 6;

	%minMass = 90;
	%maxMass = 1000;

	%dataBlock = %col.getDataBlock();

	if (%dataBlock.mass $= "" || %dataBlock.mass <= %minMass) {
		%force = %minForce;
	}
	else if (%dataBlock.mass <= %maxMass) {
		%force = mLerp(%minForce, %maxForce, mLerpFlip(%minMass, %maxMass, %dataBlock.mass));
	}
	else {
		return;
	}

	%obj.client.bottomPrint(%force, 2);

	%optimal = vectorNormalize(vectorSub(%col.getWorldBoxCenter(), %point));
	%difference = vectorScale(vectorSub(%optimal, %vector), -%directionModifier);

	%direction = vectorNormalize(vectorAdd(%vector, %difference));
	%velocity = vectorAdd(vectorScale(%vector, %force), "0 0" SPC %force / 2);

	// note: this will [s]FUCK UP[/s] do nothing with projectiles
	// gotta write code for that but lazy :>

	%col.setVelocity(%velocity);
}

// because port is old and can't remember:
// this picks up aimed at object or drops held object

function GravityGunImage::onSecondaryFire(%this, %obj, %slot) {
	if (isObject(%obj.gravityGunTarget)) {
		%obj.gravityGunTarget = "";
		cancel(%obj.gravityGunTargetUpdate);

		return;
	}

	%col = %this.getTargetObject(%obj, %slot, 50);

	if (!isObject(%col)) {
		return;
	}

	if (%col.getType() & $TypeMasks::StaticShapeObjectType) {
		return;
	}

	if (%col.getType() & $TypeMasks::FxBrickObjectType) {
		return;
	}

	if (%col.getType() & $TypeMasks::TerrainObjectType) {
		return;
	}

	%obj.gravityGunTarget = %col;
	%obj.gravityGunTargetUpdate();
}

function GravityGunImage::getTargetObject(%this, %obj, %slot, %range) {
	%point = %obj.getMuzzlePoint(%slot);
	%vector = %obj.getMuzzleVector(%slot);

	%end = vectorScale(%vector, %range);
	%end = vectorAdd(%point, %end);

	%mask = $TypeMasks::StaticShapeObjectType |
	        $TypeMasks::FxBrickObjectType |
	        $TypeMasks::TerrainObjectType |
	        $TypeMasks::VehicleObjectType |
	        $TypeMasks::PlayerObjectType;

	%ray = containerRayCast(%point, %end, %mask, %obj);
	%projectile = %this.findProjectileTarget(%obj, %slot, %range);

	if (%ray $= 0) {
		return getWord(%projectile, 0);
	}
	else if (%projectile $= 0) {
		return getWord(%ray, 0);
	}
	else {
		%d1 = vectorDist(%point, getWords(%ray, 1, 3));
		%d2 = vectorDist(%point, getWords(%projectile, 1, 3));

		if (%d2 < %d1) {
			return getWord(%projectile, 0);
		}
		else {
			return getWord(%ray, 0);
		}
	}
}

function GravityGunImage::findProjectileTarget(%this, %obj, %slot, %range) {
	%point = %obj.getMuzzlePoint(%slot);
	%vector = %obj.getMuzzleVector(%slot);

	initContainerRadiusSearch(%point, %range, $TypeMasks::ProjectileObjectType);

	while (isObject(%col = containerSearchNext())) {
		%optimal = vectorNormalize(vectorSub(%col.position, %point));
		%dotted = vectorDot(%vector, %optimal);

		if (%dotted < 0.85) {
			continue;
		}

		%distance = containerSearchRadiusCurrDist();

		if (%best $= "" || %distance < %best) {
			%best = %distance;
			%result = %col;
		}
	}

	if (%result $= "") {
		return 0;
	}

	return %result SPC %result.position;
}

function Player::gravityGunTargetUpdate(%this) {
	cancel(%this.gravityGunTargetUpdate);

	if (%this.getState() $= "Dead" || !isObject(%this.gravityGunTarget)) {
		%this.gravityGunTarget = "";
		return;
	}

	// if (%this.gravityGunTarget.getType() & $TypeMasks::PlayerObjectType) {
	// 	if (%this.gravityGunTarget.getState() $= "Dead") {
	// 		%this.gravityGunTarget = "";
	// 		return;
	// 	}
	// }

	// if (%this.gravityGunTarget.gravityGunTarget == %this || %this.getObjectMount() == %this.gravityGunTarget) {
	// 	%this.gravityGunTarget = "";
	// 	return;
	// }

	if (%this.gravityGunTarget.getType() & $TypeMasks::PlayerObjectType) {
		%current = %this.gravityGunTarget.getHackPosition();
	}
	else {
		%current = %this.gravityGunPosition.getPosition();
	}

	%optimal = vectorAdd(%this.getMuzzlePoint(0), vectorScale(%this.getMuzzleVector(0), 5));

	%difference = vectorSub(%optimal, %current);
	%velocity = vectorScale(%difference, 8);

	%this.gravityGunTarget.setVelocity(%velocity);
	%this.gravityGunTargetUpdate = %this.schedule(0, gravityGunTargetUpdate);
}

function mLerp(%a, %b, %x) {
	return %a + (%b - %a) * %x;
}

function mLerpFlip(%a, %b, %x) {
	return (%x - %a) / (%b - %a);
}

package GravityGunPackage {
	function Armor::onTrigger(%this, %obj, %slot, %bool) {
		parent::onTrigger(%this, %obj, %slot, %bool);

		if (%obj.getState() $= "Dead" || %slot != 4 || !%bool) {
			return;
		}

		for (%i = 0; %i < 4; %i++) {
			%image = %obj.getMountedImage(%i);

			if (%image == nameToID("GravityGunImage")) {
				%image.onSecondaryFire(%obj, %i);
			}
		}
	}
};

activatePackage("GravityGunPackage");
